import fastapi_users
import fastapi_users.models
import fastapi_users.db

from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base

class User(fastapi_users.models.BaseUser):
    """
    See `User model <https://fastapi-users.github.io/fastapi-users/configuration/models/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *
        from pprint import pprint

        fields = User.__fields__
        pprint(fields)
    """
    pass

class UserCreate(fastapi_users.models.BaseUserCreate):
    """
    See `UserCreate model <https://fastapi-users.github.io/fastapi-users/configuration/models/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *
        from pprint import pprint

        fields = UserCreate.__fields__
        pprint(fields)
    """
    pass

class UserUpdate(fastapi_users.models.BaseUserUpdate):
    """
    See `UserUpdate model <https://fastapi-users.github.io/fastapi-users/configuration/models/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *
        from pprint import pprint

        fields = UserUpdate.__fields__
        pprint(fields)
    """
    pass

class UserDB(User, fastapi_users.models.BaseUserDB):
    """
    See `UserDB model <https://fastapi-users.github.io/fastapi-users/configuration/models/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *
        from pprint import pprint

        fields = UserDB.__fields__
        pprint(fields)
    """
    pass

Base: DeclarativeMeta = declarative_base()
class UserTable(Base, fastapi_users.db.SQLAlchemyBaseUserTable):
    """
    See `UserTable model <https://fastapi-users.github.io/fastapi-users/configuration/databases/sqlalchemy/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *

        table = UserTable.__table__
        columns = table.c

        for c in columns:
            print(c)
    """
    pass

class UserManager(fastapi_users.BaseUserManager[UserCreate, UserDB]):
    """
    See `UserManager model <https://fastapi-users.github.io/fastapi-users/configuration/user-manager/>`_ from ``fastapi-users``.

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.models import *
        from pprint import pprint

        pprint(dir(UserManager))
    """
    user_db_model = UserDB


import databases
import os

from fastapi import Depends, Response
from fastapi_users import FastAPIUsers
from fastapi_users.authentication import CookieAuthentication, JWTAuthentication
from msdss_base_api import API
from msdss_base_database import Database
from msdss_base_dotenv import env_exists, load_env_file

from .models import *
from .tools import *

class UsersAPI(API):
    """
    Users API class for managing users.
    
    * Extends the :class:`msdss_base_api:msdss_base_api.core.API` class
    * Users will be stored in a table named ``user``

    Parameters
    ----------
    secret : str
        A secret for security encryption and protecting user data. Use a strong phrase (e.g. ``openssl rand -hex 32``).
    driver : str
        The driver name of the database connection, which are commonly ``postgresql``, ``sqlite``, ``mysql``, ``oracle`` or ``mssql``.  (see `SQLAlchemy supported databases <https://docs.sqlalchemy.org/en/14/core/engines.html#supported-databases>`_).
    user : str
        User name for the connection.
    password : str
        Password for the user.
    host : str
        Host address of the connection.
    port : str
        Port number of the connection.
    database : str
        Database name of the connection.
    users_api : :class:`fastapi_users:fastapi_users.FastAPIUsers` or None
        FastAPI Users object. See `FastAPIUsers object <https://fastapi-users.github.io/fastapi-users/configuration/routers/>`_.
        If ``None``, then one will be setup based on other parameters.
    users_api : :class:`fastapi_users:fastapi_users.FastAPIUsers` or None
        A FastAPIUsers object to setup routes with. If ``None``, one will be setup using the params.
    enable_auth_router : bool
        Whether to include the Fast API Users api route.
    enable_register_router : bool
        Whether to include the Fast API Users register route.
    enable_verify_router : bool
        Whether to include the Fast API Users verify route.
    enable_reset_password_router : bool
        Whether to include the Fast API Users reset password route.
    enable_users_router : bool
        Whether to include the Fast API Users users route.
    enable_jwt_auth : bool
        Whether to use JWT auth or not.
    enable_cookie_auth : bool
        Whether to use cookie auth or not.
    setup_startup : bool
        Whether to setup a startup event to connect databases.
    setup_shutdown : bool
        Whether to setup a shutdown event to close databases.
    auth_router_jwt_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_auth_router` using jwt auth. See `Auth router <https://fastapi-users.github.io/fastapi-users/configuration/routers/auth/>`_.
    auth_router_jwt_include_kwargs : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users auth router using jwt auth.
    auth_router_jwt_refresh : bool
        Whether to setup a token refresh route at ``auth_router_jwt_refresh_path``. This is setup according to the FastAPI Users `JWT authentication guide <https://fastapi-users.github.io/fastapi-users/configuration/authentication/jwt/>`_.
    auth_router_jwt_refresh_path : str
        Path to setup refresh token if ``auth_router_jwt_refresh`` is ``True``.
    auth_router_cookie_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_auth_router` using cookie auth. See `Auth router <https://fastapi-users.github.io/fastapi-users/configuration/routers/auth/>`_.
    auth_router_cookie_include_kwargs : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users auth router using cookie auth.
    register_router_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_register_router`. See `Register router <https://fastapi-users.github.io/fastapi-users/configuration/routers/register/>`_.
    register_router_include_kwargs : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users register router.
    register_router_superuser : bool
        Whether to limit registration control to superusers only. If ``dependencies`` are set in parameter ``register_router_include_kwargs``, then a superuser dependency will be added to the list if ``True``.
    verify_router_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_verify_router`. See `Verify router <https://fastapi-users.github.io/fastapi-users/configuration/routers/verify/>`_.
    verify_router_include_kwarg : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users verify router.
    reset_password_router_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_reset_password_router`. See `Reset password router <https://fastapi-users.github.io/fastapi-users/configuration/routers/reset/>`_.
    reset_password_router_include_kwargs : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users reset password router.
    users_router_kwargs : dict
        Keyword arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.get_users_router`. See `Users router <https://fastapi-users.github.io/fastapi-users/configuration/routers/users/>`_.
    users_router_include_kwargs : dict
        Keyword arguments passed to :meth:`fastapi:fastapi.FastAPI.include_router` for the FastAPI Users users router.
    reset_password_token_secret : str
        Secret used to secure password reset tokens. If ``None``, it will default to param ``secret``.
    verification_token_secret : str or None
        Secret used to secure verification tokens. If ``None``, it will default to param ``secret``.
    cookie_secret : str or None
        Secret used to secure cookies. If ``None``, it will default to param ``secret``.
    cookie_kwargs : dict
        Dictionary of keyword arguments passed to :class:`fastapi_users:fastapi_users.authentication.CookieAuthentication`. See `CookieAuthentication <https://fastapi-users.github.io/fastapi-users/configuration/authentication/cookie/>`_.
    jwt_secret : str or None
        Secret used to secure JWT tokens. If ``None``, it will default to param ``secret``.
    jwt_kwargs : dict
        Dictionary of keyword arguments passed to :class:`fastapi_users:fastapi_users.authentication.JWTAuthentication`.  See `JWTAuthentication <https://fastapi-users.github.io/fastapi-users/configuration/authentication/jwt/>`_.
    Base : class
        Class returned from :func:`sqlalchemy:sqlalchemy.orm.declarative_base`.
    User : :class:`msdss_users_api.models.User`
        User model for FastAPI Users. See :class:`msdss_users_api.models.User`.
    UserCreate : :class:`msdss_users_api.models.UserCreate`
        UserCreate model for FastAPI Users. See :class:`msdss_users_api.models.UserCreate`.
    UserUpdate : :class:`msdss_users_api.models.UserUpdate`
        UserUpdate model for FastAPI Users. See :class:`msdss_users_api.models.UserUpdate`.
    UserDB : :class:`msdss_users_api.models.UserDB`
        UserDB model for FastAPI Users. See :class:`msdss_users_api.models.UserDB`.
    UserTable : :class:`msdss_users_api.models.UserTable`
        UserTable model for FastAPI Users. See :class:`msdss_users_api.models.UserTable`.
    UserManager : :class:`msdss_users_api.models.UserManager`
        UserManager model for FastAPI Users. See :class:`msdss_users_api.models.UserManager`.
    database_engine : :func:`sqlalchemy:sqlalchemy.create_engine` or  None
        SQLAlchemy engine object. If ``None``, one will be created from the database connection params.
    async_database : :class:`databases:databases.Database` or None
        Async database object from ``databases``. If ``None``, one will be created from database connection params.
    get_user_db : func or None
        Function for getting the FastAPI Users DB object. If ``None``, one will be setup from params. See `FastAPI Users database config <https://fastapi-users.github.io/fastapi-users/configuration/databases/sqlalchemy/>`_.
    get_user_manager : func or None
        Function for getting the FastAPI UserManager object. If ``None``, one will be setup from params. See `FastAPI Users UserManager <https://fastapi-users.github.io/fastapi-users/configuration/user-manager/>`_.
    cookie_auth : :class:`fastapi_users:fastapi_users.authentication.CookieAuthentication` or None
        A cookie authentication object from FastAPI Users. See `CookieAuthentication <https://fastapi-users.github.io/fastapi-users/configuration/authentication/cookie/>`_.
    jwt_auth : :class:`fastapi_users:fastapi_users.authentication.JWTAuthentication` or None
        A JSON Web Token (JWT) authentication object from FastAPI Users. See `JWTAuthentication <https://fastapi-users.github.io/fastapi-users/configuration/authentication/jwt/>`_.
    load_env : bool
        Whether to load variables from a file with environmental variables at ``env_file`` or not.
    env_file : str
        The path of the file with environmental variables.
    key_path : str
        The path of the key file for the ``env_file``.
    secret_key : str
        The environmental variable name for ``secret``.
    jwt_secret_key : str
        The environmental variable name for ``jwt_secret``.
    cookie_secret_key : str
        The environmental variable name for ``cookie_secret``.
    reset_password_token_secret_key : str
        The environmental variable name for ``reset_password_token_secret``.
    verification_token_secret_key : str
        The environmental variable name for ``verification_token_secret``.
    driver_key : str
        The environmental variable name for ``driver``.
    user_key : str
        The environmental variable name for ``user``.
    password_key : str
        The environmental variable name for ``password``.
    host_key : str
        The environmental variable name for ``key``.
    port_key : str
        The environmental variable name for ``port``.
    database_key : str
        The environmental variable name for ``database``.
    log_warnings : bool
        Whether to log warnings or not.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_base_api:msdss_base_api.core.API`.

    Attributes
    ----------
    users_api : :class:`fastapi_users:fastapi_users.FastAPIUsers`
        Configured FastAPI Users object. See `FastAPIUsers object <https://fastapi-users.github.io/fastapi-users/configuration/routers/>`_.
    _users_databases : dict
        Dictionary of user database related objects:

        * ``engine`` (:func:`sqlalchemy:sqlalchemy.create_engine`): Engine from parameter ``database_engine``.
        * ``asynchronous`` (:class:`databases:databases.Database`): Async database from parameter ``async_database``.
    
    _users_functions : dict(func)
        Dictionary of user related functions from auto-config:

        * ``get_user_db`` (func): get_user_db function auto-configured or from parameter ``get_user_db``. See :func:`msdss_users_api.tools.create_user_db_func`.
        * ``get_user_manager`` (func): get_user_manager function auto-configured or from parameter ``get_user_manager``. See :func`msdss_users_api.tools.create_user_manager_func`.
    
    _users_events : dict(func)
        Dictionary of event functions registered from auto-config:

        * ``startup``: startup function registered if ``setup_startup`` is ``True``
        * ``shutdown``: shutdown function registered if ``setup_shutdown`` is ``True``
    
    _users_models : dict
        Dictionary of user models for the users API:

        * User (:class:`msdss_users_api.models.User`): see parameter ``User``.
        * UserCreate (:class:`msdss_users_api.models.UserCreate`): see parameter ``UserCreate``.
        * UserUpdate (:class:`msdss_users_api.models.UserUpdate`): see parameter ``UserUpdate``.
        * UserDB (:class:`msdss_users_api.models.UserDB`): see parameter ``UserDB``.
        * UserTable (:class:`msdss_users_api.models.UserTable`): see parameter ``UserTable``.
        * UserManager (:class:`msdss_users_api.models.UserManager`): see parameter ``UserManager``.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::
        :hide-output:

        from msdss_users_api import UsersAPI
        app = UsersAPI(
            secret='secret-phrase',
            jwt_secret='secret-phrase-02',
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Run the app with app.start()
        # Try API at http://localhost:8000/docs
        # app.start()
    """
    def __init__(
        self,
        secret='msdss-secret',
        driver='postgresql',
        user='msdss',
        password='msdss123',
        host='localhost',
        port='5432',
        database='msdss',
        users_api=None,
        enable_auth_router=True,
        enable_register_router=True,
        enable_verify_router=True,
        enable_reset_password_router=True,
        enable_users_router=True,
        enable_jwt_auth=True,
        enable_cookie_auth=True,
        setup_startup=True,
        setup_shutdown=True,
        auth_router_jwt_kwargs={},
        auth_router_jwt_include_kwargs={
            'prefix': '/auth/jwt',
            'tags': ['auth']
        },
        auth_router_jwt_refresh=True,
        auth_router_jwt_refresh_path='/refresh',
        auth_router_cookie_kwargs={},
        auth_router_cookie_include_kwargs={
            'prefix': '/auth',
            'tags': ['auth']
        },
        register_router_kwargs={},
        register_router_include_kwargs={
            'prefix': '/auth',
            'tags': ['auth']
        },
        register_router_superuser=True,
        verify_router_kwargs={},
        verify_router_include_kwargs={
            'prefix': '/auth',
            'tags': ['auth']
        },
        reset_password_router_kwargs={},
        reset_password_router_include_kwargs={
            'prefix': '/auth',
            'tags': ['auth']
        },
        users_router_kwargs={},
        users_router_include_kwargs={
            'prefix': '/users',
            'tags': ['users']
        },
        reset_password_token_secret=None,
        verification_token_secret=None,
        cookie_secret=None,
        cookie_kwargs={'lifetime_seconds': 3600},
        jwt_secret=None,
        jwt_kwargs= {
            'lifetime_seconds': 3600,
            'tokenUrl': 'auth/jwt/login'
        },
        Base=Base,
        User=User,
        UserCreate=UserCreate,
        UserUpdate=UserUpdate,
        UserDB=UserDB,
        UserTable=UserTable,
        UserManager=None,
        database_engine=None,
        async_database=None,
        get_user_db=None,
        get_user_manager=None,
        cookie_auth=None,
        jwt_auth=None,
        load_env=True,
        env_file='./.env',
        key_path=None,
        secret_key='MSDSS_USERS_SECRET',
        jwt_secret_key='MSDSS_USERS_JWT_SECRET',
        cookie_secret_key='MSDSS_USERS_COOKIE_SECRET',
        reset_password_token_secret_key='MSDSS_USERS_RESET_PASSWORD_TOKEN_SECRET',
        verification_token_secret_key='MSDSS_USERS_VERIFICATION_TOKEN_SECRET',
        driver_key='MSDSS_DATABASE_DRIVER',
        user_key='MSDSS_DATABASE_USER',
        password_key='MSDSS_DATABASE_PASSWORD',
        host_key='MSDSS_DATABASE_HOST',
        port_key='MSDSS_DATABASE_PORT',
        database_key='MSDSS_DATABASE_NAME',
        log_warnings=True,
        *args, **kwargs):
        super().__init__(*args, **kwargs)

        # (UserAPI_env) Load env vars
        has_env = env_exists(file_path=env_file, key_path=key_path)
        if load_env and has_env:
            load_env_file(file_path=env_file, key_path=key_path)
            secret = os.getenv(secret_key, secret)
            jwt_secret = os.getenv(jwt_secret_key, jwt_secret)
            cookie_secret = os.getenv(cookie_secret_key, cookie_secret)
            reset_password_token_secret = os.getenv(reset_password_token_secret_key, reset_password_token_secret_key)
            verification_token_secret = os.getenv(verification_token_secret_key, verification_token_secret)
            driver = os.getenv(driver_key, driver)
            user = os.getenv(user_key, user)
            password = os.getenv(password_key, password)
            host = os.getenv(host_key, host)
            port = os.getenv(port_key, port)
            database = os.getenv(database_key, database)
        
        # (UserAPI_warn) Warn if default secret is used
        if secret == 'msdss-secret' and log_warnings:
            self.logger.warning('Default secret was used - please change secret phrase!')

        # (UserAPI_vars) Setup vars
        jwt_secret = secret if jwt_secret is None else jwt_secret
        cookie_secret = secret if cookie_secret is None else cookie_secret

        # (UserAPI_db) Setup database connections
        db = Database(driver=driver, user=user, password=password, host=host, port=port, database=database)
        database_engine = db._connection if database_engine is None else database_engine
        async_database = databases.Database(str(database_engine.url)) if async_database is None else async_database

        # (UserAPI_auth) Setup auth for users
        jwt_auth = JWTAuthentication(secret=jwt_secret, **jwt_kwargs) if jwt_secret and jwt_auth is None else jwt_auth
        cookie_auth = CookieAuthentication(secret=cookie_secret, **cookie_kwargs) if cookie_secret and cookie_auth is None else cookie_auth
        
        # (UserAPI_auth_combine) Combine cookie and jwt auths if needed
        auth = []
        if enable_cookie_auth:
            auth.append(cookie_auth)
        if enable_jwt_auth:
            auth.append(jwt_auth)
        auth = [a for a in auth if a is not None]
        
        # (UserAPI_func) Setup required functions
        get_user_db = create_user_db_func(database_engine, async_database, sqlalchemy_base=Base, user_table_model=UserTable, user_db_model=UserDB) if get_user_db is None else get_user_db
        UserManager = create_user_manager_model(secret, reset_password_token_secret=reset_password_token_secret, verification_token_secret=verification_token_secret) if UserManager is None else UserManager
        get_user_manager = create_user_manager_func(get_user_db, UserManager) if get_user_manager is None else get_user_manager

        # (UserAPI_api) Setup users api obj
        users_api = FastAPIUsers(
            get_user_manager,
            auth,
            User,
            UserCreate,
            UserUpdate,
            UserDB
        ) if users_api is None else users_api

        # (UsersAPI_router_auth) Add auth router
        if enable_auth_router:

            # (UsersAPI_router_auth_jwt) Add jwt auth route
            if enable_jwt_auth:

                # (UsersAPI_router_auth_jwt_method) Set method of auth for route
                auth_jwt_router = users_api.get_auth_router(jwt_auth, **auth_router_jwt_kwargs)
                
                # (UsersAPI_router_auth_jwt_refresh) Create jwt auth refresh route
                if auth_router_jwt_refresh:
                    @auth_jwt_router.post(auth_router_jwt_refresh_path)
                    async def refresh_jwt(response: Response, user=Depends(users_api.current_user(active=True))):
                        return await jwt_auth.get_login_response(user, response, UserManager)

                # (UsersAPI_router_auth_jwt_add) Add jwt auth route
                self.add_router(auth_jwt_router, **auth_router_jwt_include_kwargs)

            # (UsersAPI_router_auth_cookie) Add cookie auth route
            if enable_cookie_auth:

                    # (UsersAPI_router_auth_cookie_method) Set method of auth for route
                auth_cookie_router = users_api.get_auth_router(cookie_auth, **auth_router_cookie_kwargs)

                # (UsersAPI_router_auth_cookie_add) Add auth route
                self.add_router(auth_cookie_router, **auth_router_cookie_include_kwargs)

        # (UsersAPI_router_register) Add register router
        if enable_register_router:
            if register_router_superuser:
                register_router_include_kwargs['dependencies'] = register_router_include_kwargs['dependencies'] if 'dependencies' in register_router_include_kwargs else []
                register_router_include_kwargs['dependencies'].append(Depends(users_api.current_user(superuser=True)))
            register_router = users_api.get_register_router(**register_router_kwargs)
            self.add_router(register_router, **register_router_include_kwargs)

        # (UsersAPI_router_verify) Add verify router
        if enable_verify_router:
            verify_router = users_api.get_verify_router(**verify_router_kwargs)
            self.add_router(verify_router, **verify_router_include_kwargs)
        
        # (UsersAPI_router_reset) Add reset password router
        if enable_reset_password_router:
            reset_password_router = users_api.get_reset_password_router(**reset_password_router_kwargs)
            self.add_router(reset_password_router, **reset_password_router_include_kwargs)

        # (UsersAPI_router_users) Add users router
        if enable_users_router:
            users_router = users_api.get_users_router(**users_router_kwargs)
            self.add_router(users_router, **users_router_include_kwargs)

        # (UserAPI_attr) Add attributes
        self.users_api = users_api
        self.users_database = db
        self._users_databases = dict(
            engine = database_engine,
            asynchronous = async_database
        )
        self._users_functions = dict(
            get_user_db = get_user_db,
            get_user_manager = get_user_manager
        )
        self._users_models = dict(
            User = User,
            UserCreate = UserCreate,
            UserUpdate = UserUpdate,
            UserDB = UserDB,
            UserTable = UserTable,
            UserManager = UserManager
        )
        self._users_events = {}

        # (UserAPI_events_startup) Setup app startup
        if setup_startup:
            @self.on('startup')
            async def startup():
                await self._users_databases['asynchronous'].connect()
            self._users_events['startup'] = startup

        # (UserAPI_events_shutdown) Setup app shutdown
        if setup_shutdown:
            @self.on('shutdown')
            async def shutdown():
                await self._users_databases['asynchronous'].disconnect()
            self._users_events['shutdown'] = shutdown

    def get_current_user(self, *args, **kwargs):
        """
        Get a dependency function to retrieve the current authenticated user.

        Parameters
        ----------
        *args, **kwargs
            Additional arguments passed to :meth:`fastapi_users:fastapi_users.FastAPIUsers.current_user`. See `current_user <https://fastapi-users.github.io/fastapi-users/usage/current-user/>`_.

        Return
        ------
        func
            A function to retrieve the current authenticated user. Useful for adding protected routes accessible only by authenticated users.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>

        Example
        -------
        .. jupyter-execute::
            :hide-output:

            from fastapi import Depends
            from msdss_users_api import UsersAPI
            from msdss_users_api.models import User

            app = UsersAPI(
                secret='secret-phrase',
                jwt_secret='secret-phrase-02',
                driver='postgresql',
                user='msdss',
                password='msdss123',
                host='localhost',
                port='5432',
                database='msdss'
            )

            # Get a function dependency for the current active user
            current_active_user = app.get_current_user(active=True)

            # Add a protected route
            @app.add('GET', '/protected-route')
            def protected_route(user: User = Depends(current_active_user)):
                return f'Hello, {user.email}'

            # Run the app with app.start()
            # Try API at http://localhost:8000/docs
            # app.start()
        """
        out = self.users_api.current_user(*args, **kwargs)
        return out


import contextlib
import databases
import os
import pydantic
import warnings

from fastapi import Depends
from fastapi_users.db import SQLAlchemyUserDatabase
from fastapi_users.jwt import generate_jwt
from fastapi_users.manager import UserAlreadyExists, UserNotExists
from msdss_base_database import Database
from msdss_base_dotenv import env_exists, load_env_file

from .models import *

def create_user_db_context(
    engine=None,
    async_database=None,
    load_env=True,
    env_file='./.env',
    key_path=None,
    driver_key='MSDSS_DATABASE_DRIVER',
    user_key='MSDSS_DATABASE_USER',
    password_key='MSDSS_DATABASE_PASSWORD',
    host_key='MSDSS_DATABASE_HOST',
    port_key='MSDSS_DATABASE_PORT',
    database_key='MSDSS_DATABASE_NAME',
    engine_kwargs={},
    async_database_kwargs={},
    *args, **kwargs):
    """
    Create a context manager for an auto-configured :func:`msdss_users_api.tools.create_user_db_func` function.

    Parameters
    ----------
    engine : :class:`sqlalchemy:sqlalchemy.engine.Engine` or None
        SQLAlchemy engine object to use for creating tables. If ``None``, one will be auto-configured from the ``env_file``.
    async_database : :class:`databases:databases.Database`
        The async database object to use for managing users. See `databases <https://pypi.org/project/databases/>`_. If ``None``, one will be auto-configured from the ``env_file``.
    load_env : bool
        Whether to load variables from a file with environmental variables at ``env_file`` or not.
    env_file : str
        The path of the file with environmental variables.
    key_path : str
        The path of the key file for the ``env_file``.
    driver_key : str
        The environmental variable name for ``driver``.
    user_key : str
        The environmental variable name for ``user``.
    password_key : str
        The environmental variable name for ``password``.
    host_key : str
        The environmental variable name for ``key``.
    port_key : str
        The environmental variable name for ``port``.
    database_key : str
        The environmental variable name for ``database``.
    engine_kwargs : dict
        Keyword arguments passed to :class:`msdss_base_database:msdss_base_database.core.Database` if ``engine`` is ``None``.
    async_database_kwargs : dict
        Keyword arguments passed to :class:`databases:databases.Database` if ``async_database`` is ``None``.
    *args, **kwargs
        Additional arguments passed to :func:`msdss_users_api.tools.create_user_db_func`.

    Return
    ------
    (:func:`contextlib.asynccontextmanager`, func, :class:`databases:databases.Database`, :class:`sqlalchemy:sqlalchemy.engine.Engine`)
        Returns a tuple of (in order):
        
        * Function returned from :func:`contextlib.asynccontextmanager` created from an auto-configured :func:`msdss_users_api.tools.create_user_db_func` function
        * Function for the user database dependency from ``get_user_db`` or auto-configured :func:`msdss_users_api.tools.create_user_db_func`.
        * Async database from parameter ``async_database`` or auto-configured :class:`databases:databases.Database` from env vars
        * Engine from parameter ``engine`` or auto-configured :class:`sqlalchemy:sqlalchemy.engine.Engine` from env vars

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *
        
        get_user_db_context, get_user_db, async_database, engine = create_user_db_context()
    """

    # (get_user_db_context_env) Load env vars
    has_env = env_exists(file_path=env_file, key_path=key_path)
    if load_env and has_env:
        engine_kwargs['driver'] = os.getenv(driver_key, 'postgresql')
        engine_kwargs['user'] = os.getenv(user_key, 'msdss')
        engine_kwargs['password'] = os.getenv(password_key, 'msdss123')
        engine_kwargs['host'] = os.getenv(host_key, 'localhost')
        engine_kwargs['port'] = os.getenv(port_key, '5432')
        engine_kwargs['database'] = os.getenv(database_key, 'msdss')
    
    # (create_user_db_func_db) Create databases
    engine = engine if engine else Database(**engine_kwargs)._connection
    async_database = async_database if async_database else databases.Database(str(engine.url), **async_database_kwargs)
    
    # (get_user_db_context_return) Return user db context
    get_user_db = create_user_db_func(engine=engine, async_database=async_database, *args, **kwargs)
    out = (contextlib.asynccontextmanager(get_user_db), get_user_db, async_database, engine)
    return out

def create_user_db_func(
    engine,
    async_database,
    sqlalchemy_base=Base,
    user_table_model=UserTable,
    user_db_model=UserDB):
    """
    Create a function to return the the database adapter dependency.

    See `SQLAlchemy configuration <https://fastapi-users.github.io/fastapi-users/configuration/databases/sqlalchemy/>`_ for FastAPI Users,

    Parameters
    ----------
    engine : :class:`sqlalchemy:sqlalchemy.engine.Engine`
        SQLAlchemy engine object to use for creating tables.
    async_database : :class:`databases:databases.Database`
        The async database object to use for managing users. See `databases <https://pypi.org/project/databases/>`_.
    sqlalchemy_base : :func:`sqlalchemy:sqlalchemy.orm.declarative_base`
        The base class from sqlalchemy. See :func:`sqlalchemy:sqlalchemy.orm.declarative_base`.
    user_table_model : :class:`msdss_users_api.models.UserTable`
        The user table model to use for the database dependency. See :class:`msdss_users_api.models.UserTable`.
    user_db_model : :class:`msdss_users_api.models.UserDB`
        The user database model for the database dependency. See :class:`msdss_users_api.models.UserDB`.

    Return
    ------
    func
        A function yielding a :class:`fastapi_users:fastapi_users.db.SQLAlchemyUserDatabase`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::

        import databases

        from msdss_base_database import Database
        from msdss_users_api.tools import create_user_db_func

        # Create a database object
        db = Database()

        # Get the engine object from the database
        engine = db._connection

        # Get the async database
        async_database = databases.Database(str(engine.url))

        # Get the user db func
        get_user_db = create_user_db_func(engine, async_database)
    """

    # (create_user_db_func_table) Create user table in database
    sqlalchemy_base.metadata.create_all(engine)
    table = user_table_model.__table__

    # (create_user_db_func_return) Return the get_user_db function
    async def out():
        yield SQLAlchemyUserDatabase(user_db_model, async_database, table)
    return out

def create_user_manager_context(
    secret='msdss-secret',
    get_user_db=None,
    reset_password_token_secret=None,
    verification_token_secret=None,
    user_manager_model=None,
    load_env=True,
    env_file='./.env',
    key_path=None,
    secret_key='MSDSS_USERS_SECRET',
    reset_password_token_secret_key='MSDSS_USERS_RESET_PASSWORD_TOKEN_SECRET',
    verification_token_secret_key='MSDSS_USERS_VERIFICATION_TOKEN_SECRET',
    create_user_manager_model_kwargs={}):
    """
    Create a context manager for an auto-configured :func:`msdss_users_api.tools.create_user_manager_func` function.

    Parameters
    ----------
    secret : str
        A secret for security encryption and protecting user data. Use a strong phrase (e.g. ``openssl rand -hex 32``).
    get_user_db : func
        Function for the user database dependency. See :func:`msdss_users_api.tools.create_user_db_func`.
    reset_password_token_secret : str
        Secret used to secure password reset tokens. If ``None``, it will default to param ``secret``.
    verification_token_secret : str or None
        Secret used to secure verification tokens. If ``None``, it will default to param ``secret``.
    user_manager_model : :class:`msdss_users_api.models.UserManager`
        The user manager model from FastAPI Users. See :class:`msdss_users_api.models.UserManager` and :func:`msdss_users_api.tools.create_user_manager_model`.
    load_env : bool
        Whether to load variables from a file with environmental variables at ``env_file`` or not.
    env_file : str
        The path of the file with environmental variables.
    key_path : str
        The path of the key file for the ``env_file``.
    secret_key : str
        The environmental variable name for ``secret``.
    reset_password_token_secret_key : str
        The environmental variable name for ``reset_password_token_secret``.
    verification_token_secret_key : str
        The environmental variable name for ``verification_token_secret``.
    create_user_manager_model_kwargs : dict
        Keyword arguments passed to :func:`msdss_users_api.tools.create_user_manager_model`

    Return
    ------
    :func:`contextlib.asynccontextmanager`
        Function returned from :func:`contextlib.asynccontextmanager` created from an auto-configured :func:`msdss_users_api.tools.create_user_manager_func` function.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *
        
        get_user_manager_context = create_user_manager_context('msdss-secret')
    """

    # (get_user_manager_warn) Warn if default secret is used
    if secret == 'msdss-secret':
        warnings.warn('Default secret was used - please change secret phrase!')
    
    # (get_user_manager_context_env) Load env vars
    has_env = env_exists(file_path=env_file, key_path=key_path)
    if load_env and has_env:
        load_env_file(file_path=env_file, key_path=key_path)
        secret = os.getenv(secret_key, secret)
        reset_password_token_secret = os.getenv(reset_password_token_secret_key, reset_password_token_secret_key)
        verification_token_secret = os.getenv(verification_token_secret_key, verification_token_secret)
    
    # (get_user_manager_vars) Set default vars
    reset_password_token_secret = reset_password_token_secret if reset_password_token_secret else secret
    verification_token_secret = verification_token_secret if verification_token_secret else secret

    # (get_user_manager_context_func) Create user manager func
    user_manager_model = user_manager_model if user_manager_model else create_user_manager_model(
        secret,
        reset_password_token_secret=reset_password_token_secret,
        verification_token_secret=verification_token_secret,
        **create_user_manager_model_kwargs)
    get_user_manager = create_user_manager_func(get_user_db, user_manager_model)

    # (get_user_manager_context_return) Return user manager context
    out = contextlib.asynccontextmanager(get_user_manager)
    return out

def create_user_manager_func(get_user_db, user_manager_model=UserManager):
    """
    Create a function to return the the user manager class.

    See `UserManager configuration <https://fastapi-users.github.io/fastapi-users/configuration/user-manager/>`_ for FastAPI Users,

    Parameters
    ----------
    get_user_db : func
        Function for the user database dependency. See :func:`msdss_users_api.tools.create_user_db_func`.
    user_manager_model : :class:`msdss_users_api.models.UserManager`
        The user manager model from FastAPI Users. See :class:`msdss_users_api.models.UserManager` and :func:`msdss_users_api.tools.create_user_manager_model`.

    Return
    ------
    func
        A function yielding a configured :class:`msdss_users_api.models.UserManager`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::

        import databases

        from msdss_base_database import Database
        from msdss_users_api.tools import *

        # Create a database object
        db = Database()

        # Get the engine object from the database
        engine = db._connection

        # Get the async database
        async_database = databases.Database(str(engine.url))

        # Get the user db func
        get_user_db = create_user_db_func(engine, async_database)

        # Create a user manager model
        UserManager = create_user_manager_model('msdss-secret')

        # Get the user manager func
        get_user_manager = create_user_manager_func(get_user_db, UserManager)
    """
    async def out(user_db=Depends(get_user_db)):
        yield user_manager_model(user_db)
    return out

def create_user_manager_model(
    secret,
    reset_password_token_secret=None,
    verification_token_secret=None,
    base_user_manager_model=UserManager,
    *args, **kwargs):
    """
    Create a function to return the the user manager class.

    See `UserManager configuration <https://fastapi-users.github.io/fastapi-users/configuration/user-manager/>`_ for FastAPI Users,

    Parameters
    ----------
    secret : str
        Secret to use for reset password tokens and verification tokens encryption.
    reset_password_token_secret : str or None
        Secret to use for reset password token encryption. If ``None``, then it will default to param ``secret``.
    verification_token_secret : str or None
        Secret to use for verification tokens encryption. If ``None``, then it will default to param ``secret``.
    base_user_manager_model : :class:`msdss_users_api.models.UserManager`
        The user manager model from FastAPI Users. See :class:`msdss_users_api.models.UserManager`.
    *args, **kwargs
        Additional arguments passed to :func:`pydantic:pydantic.create_model`. See `pydantic dynamic model creation <https://pydantic-docs.helpmanual.io/usage/models/#dynamic-model-creation>`_.

    Return
    ------
    :class:`msdss_users_api.models.UserManager`
       A configured :class:`msdss_users_api.models.UserManager`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::

        from msdss_users_api.tools import create_user_manager_model

        UserManager = create_user_manager_model('msdss-secret')
    """

    # (create_user_manager_model_vars) Set default vars
    reset_password_token_secret = secret if reset_password_token_secret is None else reset_password_token_secret
    verification_token_secret = secret if verification_token_secret is None else verification_token_secret

    # (create_user_manager_model_return) Returns created user manager model
    out = pydantic.create_model(
        'UserManager',
        reset_password_token_secret=reset_password_token_secret,
        verification_token_secret=verification_token_secret,
        __base__=base_user_manager_model, *args, **kwargs)
    return out

async def delete_user(email, create_user_db_context_kwargs={}, create_user_manager_context_kwargs={}):
    """
    Delete a user.

    Parameters
    ----------
    email : str
        Email for the user.
    create_user_db_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_db_context`.
    create_user_manager_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_manager_context`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *

        await register_user('test@example.com', 'msdss123')
        await delete_user('test@example.com')
    """

    # (delete_user_context) Get db and manager context functions
    get_user_db_context, get_user_db, async_database, engine = create_user_db_context(**create_user_db_context_kwargs)
    get_user_manager_context = create_user_manager_context(get_user_db=get_user_db, **create_user_manager_context_kwargs)

    # (delete_user_run) Run delete user function
    try:
        async with get_user_db_context() as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                await async_database.connect()
                user = await user_manager.get_by_email(email)
                await user_manager.delete(user)
                print(f'User deleted {email}')
    except UserNotExists:
        print(f'User {email} does not exist')
    finally:
        await async_database.disconnect()

async def get_user(email, show=False, include_hashed_password=False, create_user_db_context_kwargs={}, create_user_manager_context_kwargs={}):
    """
    Get attributes for a user.

    Parameters
    ----------
    email : str
        Email for the user.
    show : bool
        Whether to print user attributes or not.
    include_hashed_password : bool
        Whether to include the ``hashed_password`` attribute.
    create_user_db_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_db_context`.
    create_user_manager_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_manager_context`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Return
    ------
    :class:`msdss_users.models.User`
        A user model with attributes for the specified user with ``email``.

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import register_user, get_user

        await register_user('test@example.com', 'msdss123')
        user = await get_user('test@example.com', show=True)

    """

    # (get_user_context) Get db and manager context functions
    get_user_db_context, get_user_db, async_database, engine = create_user_db_context(**create_user_db_context_kwargs)
    get_user_manager_context = create_user_manager_context(get_user_db=get_user_db, **create_user_manager_context_kwargs)

    # (get_user_run) Run get user function
    try:
        async with get_user_db_context() as user_db:
            async with get_user_manager_context(user_db) as user_manager:

                # (get_user_run_get) Connect and get the user from db
                await async_database.connect()
                out = await user_manager.get_by_email(email)

                # (get_user_run_hash) Remove hashed password if needed
                if not include_hashed_password:
                    del out.hashed_password
                
                # (get_user_run_show) Print user attributes
                if show:
                    for k, v in out.dict().items():
                        print(f'{k}: {v}')
                return out
    except UserNotExists:
        print(f'User {email} does not exist')
    finally:
        await async_database.disconnect()

async def register_user(
    email,
    password,
    superuser=False,
    create_user_db_context_kwargs={},
    create_user_manager_context_kwargs={},
    *args, **kwargs):
    """
    Register a user.

    Parameters
    ----------
    email : str
        Email for the user.
    password : str
        Password for the user.
    superuser : bool
        Whether the user is a superuser or not.
    create_user_db_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_db_context`.
    create_user_manager_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_manager_context`.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_users_api.models.UserCreate`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *
        
        await register_user('test@example.com', 'msdss123')
    """
    
    # (register_user_context) Get db and manager context functions
    get_user_db_context, get_user_db, async_database, engine = create_user_db_context(**create_user_db_context_kwargs)
    get_user_manager_context = create_user_manager_context(get_user_db=get_user_db, **create_user_manager_context_kwargs)

    # (register_user_run) Run create user function
    try:
        async with get_user_db_context() as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                await async_database.connect()
                await user_manager.create(
                    UserCreate(
                        email=email,
                        password=password,
                        is_superuser=superuser,
                        *args, **kwargs))
                print(f'User created {email}')
    except UserAlreadyExists:
        print(f'User {email} already exists')
    finally:
        await async_database.disconnect()

async def reset_user_password(email, password, create_user_db_context_kwargs={}, create_user_manager_context_kwargs={}):
    """
    Reset password for a user.

    Parameters
    ----------
    email : str
        Email for the user.
    password : str
        New password for the user.
    create_user_db_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_db_context`.
    create_user_manager_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_manager_context`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *

        await register_user('test@example.com', 'msdss123')
        await reset_user_password('test@example.com', 'msdss321')
    """

    # (reset_user_password_context) Get db and manager context functions
    get_user_db_context, get_user_db, async_database, engine = create_user_db_context(**create_user_db_context_kwargs)
    get_user_manager_context = create_user_manager_context(get_user_db=get_user_db, **create_user_manager_context_kwargs)

    # (reset_user_password_run) Run password reset user function
    try:
        async with get_user_db_context() as user_db:
            async with get_user_manager_context(user_db) as user_manager:

                # (reset_user_password_run_user) Get user by email
                await async_database.connect()
                user = await user_manager.get_by_email(email)

                # (reset_user_password_run_token) Get forgot password token
                token_data = {
                    "user_id": str(user.id),
                    "aud": user_manager.reset_password_token_audience,
                }
                token = generate_jwt(
                    token_data,
                    user_manager.reset_password_token_secret,
                    user_manager.reset_password_token_lifetime_seconds,
                )

                # (reset_user_password_run_reset) Reset password with token
                await user_manager.reset_password(token, password)
                print(f'User password reset {email}')
    except UserNotExists:
        print(f'User {email} does not exist')
    finally:
        await async_database.disconnect()

async def update_user(
    email,
    create_user_db_context_kwargs={},
    create_user_manager_context_kwargs={},
    *args, **kwargs):
    """
    Update a user.

    Parameters
    ----------
    email : str
        Email for the user.
    create_user_db_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_db_context`.
    create_user_manager_context_kwargs : dict
        Arguments passed to :class:`msdss_users_api.tools.create_user_manager_context`.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_users_api.models.UserUpdate`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::

        from msdss_users_api.tools import *
        
        # Create a test user
        await register_user('test@example.com', 'msdss123')
        print('\\nbefore_update:\\n')
        user = await get_user('test@example.com', show=True)
        
        # Update the test user
        await update_user('test@example.com', is_verified=True)
        print('\\nafter_update:\\n')
        user = await get_user('test@example.com', show=True)
    """
    
    # (register_user_context) Get db and manager context functions
    get_user_db_context, get_user_db, async_database, engine = create_user_db_context(**create_user_db_context_kwargs)
    get_user_manager_context = create_user_manager_context(get_user_db=get_user_db, **create_user_manager_context_kwargs)

    # (register_user_run) Run create user function
    try:
        async with get_user_db_context() as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                await async_database.connect()
                user = await user_manager.get_by_email(email)
                await user_manager.update(
                    UserUpdate(
                        email=email,
                        *args, **kwargs),
                    user
                )
                print(f'User updated {email}')
    except UserNotExists:
        print(f'User {email} does not exist')
    finally:
        await async_database.disconnect()


import argparse
import asyncio

from getpass import getpass

from .tools import *

def _get_parser():
    """
    Builds an ``argparse`` parser for the ``msdss-users`` command line tool.

    Returns
    -------
    :class:`argparse.ArgumentParser`
        An ``argparse`` parser for ``msdss-users``.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. jupyter-execute::
        :hide-output:

        from msdss_users_api.cli import _get_parser

        parser = _get_parser()
        parser.print_help()
    """

    # (_get_parser_parsers) Create main parser and sub parsers
    parser = argparse.ArgumentParser(description='Manages users with a database')
    subparsers = parser.add_subparsers(title='commands', dest='command')
    
    # (_get_parser_register) Add register command
    register_parser = subparsers.add_parser('register', help='register a user')
    register_parser.add_argument('email', type=str, nargs='?', help='email for user')
    register_parser.add_argument('password', type=str, nargs='?', help='password for user')
    register_parser.add_argument('--superuser', dest='superuser', action='store_true', help='register a superuser')
    register_parser.set_defaults(superuser=False)

    # (_get_parser_get) Add get command
    get_parser = subparsers.add_parser('get', help='get user attributes')
    get_parser.add_argument('email', type=str, help='email for user')

    # (_get_parser_delete) Add delete command
    delete_parser = subparsers.add_parser('delete', help='delete a user')
    delete_parser.add_argument('email', type=str, help='email of user to delete')

    # (_get_parser_reset) Add reset command
    reset_parser = subparsers.add_parser('reset', help='reset user password')
    reset_parser.add_argument('email', type=str, help='email of user to reset')
    reset_parser.add_argument('password', type=str, nargs='?', help='new password to use')

    # (_get_parser_update) Add update command
    update_parser = subparsers.add_parser('update', help='update a user\'s attribute')
    update_parser.add_argument('email', type=str, help='email of user')
    update_parser.add_argument('--is_active', type=bool, default=None, help='set is_active attribute')
    update_parser.add_argument('--is_superuser', type=bool, default=None, help='set is_superuser attribute')
    update_parser.add_argument('--is_verified', type=bool, default=None, help='set is_verified attribute')

    # (_get_parser_file_key) Add file and key arguments to all commands
    for p in [parser, register_parser, delete_parser, update_parser, get_parser, reset_parser]:
        p.add_argument('--env_file', type=str, default='./.env', help='path of .env file')
        p.add_argument('--key_path', type=str, default=None, help='path of key file')
    
    # (_get_parser_out) Return the parser
    out = parser
    return out

def _prompt_password():
    """
    Prompts the user for a password.

    Returns
    -------
    str
        Password provided by user input.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    .. code:: python

        from msdss_users_api.cli import _prompt_password
        _prompt_password()
    """
    password = getpass()
    verify = getpass('Verify password: ')
    if not password == verify:
        raise ValueError('Passwords do not match')
    return password

def run():
    """
    Runs the ``msdss-users`` command.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    >>> msdss-users --help

    .. jupyter-execute::
        :hide-code:

        from msdss_users_api.cli import _get_parser

        parser = _get_parser()
        parser.print_help()

    Create a user interactively:

    >>> msdss-users register

    Get user attributes:

    >>> msdss-users get test@example.com

    Update user attributes:

    >>> msdss-users update test@example.com --is_verified True

    Reset user password:

    >>> msdss-users reset test@example.com

    Delete a user:

    >>> msdss-users delete test@example.com
    """

    # (run_kwargs) Get arguments and command
    parser = _get_parser()
    kwargs = vars(parser.parse_args())
    command = kwargs.pop('command')

    # (run_env) Get env files and paths
    env_kwargs = dict(
        env_file=kwargs.pop('env_file'),
        key_path=kwargs.pop('key_path')
    )

    # (run_command) Run commands
    if command == 'register':
        kwargs['email'] = kwargs['email'] if kwargs['email'] else input('Email: ')
        kwargs['password'] = kwargs['password'] if kwargs['password'] else _prompt_password()
        asyncio.run(register_user(
            create_user_db_context_kwargs=env_kwargs,
            create_user_manager_context_kwargs=env_kwargs,
            **kwargs
        ))
    elif command == 'get':
        asyncio.run(get_user(
            show=True,
            create_user_db_context_kwargs=env_kwargs,
            create_user_manager_context_kwargs=env_kwargs,
            **kwargs
        ))
    elif command == 'delete':
        asyncio.run(delete_user(
            create_user_db_context_kwargs=env_kwargs,
            create_user_manager_context_kwargs=env_kwargs,
            **kwargs
        ))
    elif command == 'reset':
        kwargs['password'] = kwargs['password'] if kwargs['password'] else _prompt_password()
        asyncio.run(reset_user_password(
            create_user_db_context_kwargs=env_kwargs,
            create_user_manager_context_kwargs=env_kwargs,
            **kwargs
        ))
    elif command == 'update':
        kwargs = {k:v for k, v in kwargs.items() if v}
        asyncio.run(update_user(
            create_user_db_context_kwargs=env_kwargs,
            create_user_manager_context_kwargs=env_kwargs,
            **kwargs
        ))